from __future__ import annotations

import math
from typing import Tuple, Type

import jax.numpy as jnp
from jax.nn import softplus
from tjax import RealArray, RealNumeric, Shape
from tjax.dataclasses import dataclass

from ..exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import VectorSupport, distribution_parameter
from ..tools import inverse_softplus, ive, log_ive

__all__ = ['VonMisesFisherNP', 'VonMisesFisherEP']


@dataclass
class VonMisesFisherNP(NaturalParametrization['VonMisesFisherEP', RealArray]):
    mean_times_concentration: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_times_concentration.shape[:-1]

    def log_normalizer(self) -> RealArray:
        half_k = self.mean_times_concentration.shape[-1] * 0.5
        kappa = jnp.linalg.norm(self.mean_times_concentration, 2, axis=-1)
        return (kappa
                - (half_k - 1.0) * jnp.log(kappa)
                + half_k * jnp.log(2.0 * math.pi)
                + log_ive(half_k - 1.0, kappa))

    def to_exp(self) -> VonMisesFisherEP:
        q = self.mean_times_concentration
        kappa = jnp.linalg.norm(q, 2, axis=-1, keepdims=True)
        return VonMisesFisherEP(
            jnp.where(kappa == 0.0,
                      q,
                      q * (_a_k(q.shape[-1], kappa) / kappa)))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(self.shape)

    def sufficient_statistics(self, x: RealArray) -> VonMisesFisherEP:
        return VonMisesFisherEP(x)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean_times_concentration.shape[-1]

    def to_kappa_angle(self) -> Tuple[RealArray, RealArray]:
        if self.mean_times_concentration.shape[-1] != 2:
            raise ValueError
        kappa = jnp.linalg.norm(self.mean_times_concentration, axis=-1)
        angle = jnp.where(kappa == 0.0,
                          0.0,
                          jnp.arctan2(self.mean_times_concentration[..., 1],
                                      self.mean_times_concentration[..., 0]))
        return kappa, angle


@dataclass
class VonMisesFisherEP(ExpToNat[VonMisesFisherNP, RealArray]):
    mean: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[VonMisesFisherNP]:
        return VonMisesFisherNP

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def initial_search_parameters(self) -> RealArray:
        k = self.mean.shape[-1]
        mu = jnp.linalg.norm(self.mean, 2, axis=-1)
        # 0 <= mu <= 1.0
        initial_kappa = jnp.where(mu == 1.0,
                                  jnp.inf,
                                  (mu * k - mu ** 3) / (1.0 - mu ** 2))
        return inverse_softplus(initial_kappa)

    def search_to_natural(self, search_parameters: RealArray) -> VonMisesFisherNP:
        kappa = softplus(search_parameters)
        mu = jnp.linalg.norm(self.mean, 2, axis=-1)
        q = self.mean * (kappa / mu)[..., jnp.newaxis]
        return VonMisesFisherNP(q)

    def search_gradient(self, search_parameters: RealArray) -> RealArray:
        k = self.mean.shape[-1]
        kappa = softplus(search_parameters)
        mu = jnp.linalg.norm(self.mean, 2, axis=-1)
        return _a_k(k, kappa) - mu

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]


# Private functions --------------------------------------------------------------------------------
def _a_k(k: RealNumeric, kappa: RealNumeric) -> RealNumeric:
    half_k = k * 0.5
    return ive(half_k, kappa) / ive(half_k - 1.0, kappa)



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..samplable import Samplable
from .dirichlet_common import DirichletCommonEP, DirichletCommonNP

__all__ = ['BetaNP', 'BetaEP']


@dataclass
class BetaNP(DirichletCommonNP['BetaEP'], Samplable):
    """
    The Beta distribution.

    The best way to interpret the parameters of the beta distribution are that an observation x in
    [0, 1] represents the Bernoulli probability that outcome 0 (out of {0, 1}) is realized.  In this
    way, the Beta class coincides with a special case of the Dirichlet class.
    """
    # Overridden methods ---------------------------------------------------------------------------
    def to_exp(self) -> BetaEP:
        return BetaEP(self._exp_helper())

    def sufficient_statistics(self, x: RealArray) -> BetaEP:
        return BetaEP(jnp.stack([jnp.log(x), jnp.log(1.0 - x)], axis=-1))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        return jax.random.beta(rng.key,
                               1.0 + self.alpha_minus_one[..., 0],
                               1.0 + self.alpha_minus_one[..., 1],
                               shape)  # type: ignore


@dataclass
class BetaEP(DirichletCommonEP[BetaNP]):
    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[BetaNP]:
        return BetaNP

    def search_to_natural(self, search_parameters: RealArray) -> BetaNP:
        return BetaNP(self._transform_nat_helper(search_parameters))



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable

__all__ = ['WeibullNP', 'WeibullEP']


@dataclass
class WeibullNP(NaturalParametrization['WeibullEP', RealArray]):
    concentration: RealArray = distribution_parameter(ScalarSupport(), fixed=True)
    # eta = -scale^-concentration
    eta: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.eta.shape

    def log_normalizer(self) -> RealArray:
        return -jnp.log(-self.eta) - jnp.log(self.concentration)

    def to_exp(self) -> WeibullEP:
        return WeibullEP(self.concentration, -1.0 / self.eta)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return (self.concentration - 1.0) * jnp.log(x)

    def sufficient_statistics(self, x: RealArray) -> WeibullEP:
        return WeibullEP(jnp.broadcast_to(self.concentration, x.shape), x ** self.concentration)


@dataclass
class WeibullEP(ExpectationParametrization[WeibullNP], Samplable):
    concentration: RealArray = distribution_parameter(ScalarSupport(), fixed=True)
    # chi = scale^concentration
    chi: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.chi.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[WeibullNP]:
        return WeibullNP

    def to_nat(self) -> WeibullNP:
        return WeibullNP(self.concentration, -1.0 / self.chi)

    def expected_carrier_measure(self) -> RealArray:
        k = self.concentration
        lambda_ = self.chi ** (1.0 / k)
        return (k - 1.0) * jnp.log(lambda_) - jnp.euler_gamma * (1.0 - 1.0 / k)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        lambda_ = self.chi ** (1.0 / self.concentration)
        return jax.random.weibull_min(rng.key, lambda_, self.concentration, shape)  # type: ignore



from ._src import (ExpectationParametrization, HasConjugatePrior, NaturalParametrization, Samplable,
                   ScalarSupport, SquareMatrixSupport, Support, SymmetricMatrixSupport,
                   VectorSupport)
from ._src.distributions import (BernoulliEP, BernoulliNP, BetaEP, BetaNP, ChiEP, ChiNP,
                                 ChiSquareEP, ChiSquareNP, ComplexCircularlySymmetricNormalEP,
                                 ComplexCircularlySymmetricNormalNP,
                                 ComplexMultivariateUnitNormalEP, ComplexMultivariateUnitNormalNP,
                                 ComplexNormalEP, ComplexNormalNP, DirichletEP, DirichletNP,
                                 ExponentialEP, ExponentialNP, GammaEP, GammaNP, GeometricEP,
                                 GeometricNP, IsotropicNormalEP, IsotropicNormalNP, LogarithmicEP,
                                 LogarithmicNP, MultinomialEP, MultinomialNP,
                                 MultivariateDiagonalNormalEP, MultivariateDiagonalNormalNP,
                                 MultivariateDiagonalNormalVP, MultivariateNormalEP,
                                 MultivariateNormalNP, MultivariateNormalVP,
                                 MultivariateUnitNormalEP, MultivariateUnitNormalNP,
                                 NegativeBinomialEP, NegativeBinomialNP, NormalEP, NormalNP,
                                 PoissonEP, PoissonNP, RayleighEP, RayleighNP, VonMisesFisherEP,
                                 VonMisesFisherNP, WeibullEP, WeibullNP)
from ._src.scipy_replacement import (ScipyComplexMultivariateNormal, ScipyComplexNormal,
                                     ScipyDirichlet, ScipyMultivariateNormal, ScipyVonMises)



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
import numpy as np
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable

__all__ = ['NormalNP', 'NormalEP']


@dataclass
class NormalNP(NaturalParametrization['NormalEP', RealArray]):
    mean_times_precision: RealArray = distribution_parameter(ScalarSupport())
    negative_half_precision: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_times_precision.shape

    def log_normalizer(self) -> RealArray:
        return (-jnp.square(self.mean_times_precision) / (4.0 * self.negative_half_precision)
                + 0.5 * jnp.log(-np.pi / self.negative_half_precision))

    def to_exp(self) -> NormalEP:
        mean = -self.mean_times_precision / (2.0 * self.negative_half_precision)
        second_moment = jnp.square(mean) - 0.5 / self.negative_half_precision
        return NormalEP(mean, second_moment)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sufficient_statistics(self, x: RealArray) -> NormalEP:
        return NormalEP(x, jnp.square(x))


@dataclass
class NormalEP(ExpectationParametrization[NormalNP], Samplable):
    mean: RealArray = distribution_parameter(ScalarSupport())
    second_moment: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[NormalNP]:
        return NormalNP

    def to_nat(self) -> NormalNP:
        return NormalNP(self.mean / self.variance(), -0.5 / self.variance())

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        deviation = jnp.sqrt(self.variance())
        return jax.random.normal(rng.key, shape) * deviation + self.mean

    # New methods ----------------------------------------------------------------------------------
    def variance(self) -> RealArray:
        return self.second_moment - jnp.square(self.mean)



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from jax.nn import softplus
from tjax import RealArray, Shape
from tjax.dataclasses import dataclass

from ..exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter

__all__ = ['LogarithmicNP', 'LogarithmicEP']


@dataclass
class LogarithmicNP(NaturalParametrization['LogarithmicEP', RealArray]):
    log_probability: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.log_probability.shape

    def log_normalizer(self) -> RealArray:
        return jnp.log(-jnp.log1p(-jnp.exp(self.log_probability)))

    def to_exp(self) -> LogarithmicEP:
        probability = jnp.exp(self.log_probability)
        chi = jnp.where(self.log_probability < -50.0, 1.0,
                        jnp.where(self.log_probability > -1e-7, jnp.inf,
                                  probability / (jnp.expm1(self.log_probability)
                                                 * jnp.log1p(-probability))))
        return LogarithmicEP(chi)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return -jnp.log(x)

    def sufficient_statistics(self, x: RealArray) -> LogarithmicEP:
        return LogarithmicEP(x)


@dataclass
class LogarithmicEP(ExpToNat[LogarithmicNP, RealArray]):
    chi: RealArray = distribution_parameter(ScalarSupport())  # - odds / log(1-p)

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.chi.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[LogarithmicNP]:
        return LogarithmicNP

    # The expected_carrier_measure is unknown.

    def initial_search_parameters(self) -> RealArray:
        return jnp.zeros(self.chi.shape)

    def search_to_natural(self, search_parameters: RealArray) -> LogarithmicNP:
        # Run Newton's method on the whole real line.
        return LogarithmicNP(-softplus(-search_parameters))

    def search_gradient(self, search_parameters: RealArray) -> RealArray:
        return self._natural_gradient(self.search_to_natural(search_parameters)).log_probability

    # Overridden methods ---------------------------------------------------------------------------
    def to_nat(self) -> LogarithmicNP:
        z = super().to_nat()
        return LogarithmicNP(jnp.where(self.chi < 1.0,
                                       jnp.nan,
                                       jnp.where(self.chi == 1.0,
                                                 jnp.inf,
                                                 z.log_probability)))



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
import numpy as np
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter
from ...samplable import Samplable

__all__ = ['MultivariateDiagonalNormalNP', 'MultivariateDiagonalNormalEP',
           'MultivariateDiagonalNormalVP']


@dataclass
class MultivariateDiagonalNormalNP(NaturalParametrization['MultivariateDiagonalNormalEP',
                                                          RealArray]):
    mean_times_precision: RealArray = distribution_parameter(VectorSupport())
    negative_half_precision: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    def log_normalizer(self) -> RealArray:
        components = (-jnp.square(self.mean_times_precision) / (4.0 * self.negative_half_precision)
                      + 0.5 * jnp.log(-np.pi / self.negative_half_precision))
        return jnp.sum(components, axis=-1)

    def to_exp(self) -> MultivariateDiagonalNormalEP:
        mean = -self.mean_times_precision / (2.0 * self.negative_half_precision)
        second_moment = jnp.square(mean) - 0.5 / self.negative_half_precision
        return MultivariateDiagonalNormalEP(mean, second_moment)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape[:-1])

    def sufficient_statistics(self, x: RealArray) -> MultivariateDiagonalNormalEP:
        return MultivariateDiagonalNormalEP(x, jnp.square(x))

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateDiagonalNormalEP(ExpectationParametrization[MultivariateDiagonalNormalNP],
                                   Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())
    second_moment: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[MultivariateDiagonalNormalNP]:
        return MultivariateDiagonalNormalNP

    def to_nat(self) -> MultivariateDiagonalNormalNP:
        return MultivariateDiagonalNormalNP(self.mean / self.variance(), -0.5 / self.variance())

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        return self.to_variance_parametrization().sample(rng, shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> RealArray:
        return self.second_moment - jnp.square(self.mean)

    def to_variance_parametrization(self) -> MultivariateDiagonalNormalVP:
        return MultivariateDiagonalNormalVP(self.mean, self.variance())


@dataclass
class MultivariateDiagonalNormalVP(Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())
    variance: RealArray = distribution_parameter(VectorSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        deviation = jnp.sqrt(self.variance)
        return jax.random.normal(rng.key, shape) * deviation + self.mean

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def to_exp(self) -> MultivariateDiagonalNormalEP:
        second_moment = self.variance + jnp.square(self.mean)
        return MultivariateDiagonalNormalEP(self.mean, second_moment)



from .conjugate_prior import *
from .distributions import *
from .expectation_parametrization import *
from .natural_parametrization import *
from .parameter import *
from .samplable import *
from .scipy_replacement import *
from .tools import *



from __future__ import annotations

from typing import Any, List

import numpy as np
import numpy.typing as npt
from numpy.random import Generator
from tjax import ComplexArray, RealArray, Shape

__all__: List[str] = []


class ShapedDistribution:
    """
    Allow a distributions with shape.
    """
    def __init__(self,
                 shape: Shape,
                 rvs_shape: Shape,
                 rvs_dtype: np.dtype[Any],
                 objects: npt.NDArray[np.object_]):
        self.shape = shape
        self.rvs_shape = rvs_shape
        self.rvs_dtype = rvs_dtype
        self.real_dtype: np.dtype[Any] = np.zeros(0, dtype=rvs_dtype).real.dtype
        self.objects = objects

    def rvs(self, size: Shape = None, random_state: Generator = None) -> RealArray:
        if size is None:
            size = ()
        elif isinstance(size, int):
            size = (size,)
        retval = np.empty(self.shape + size + self.rvs_shape,
                          dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].rvs(size=size, random_state=random_state)
        return retval

    def pdf(self, x: ComplexArray) -> RealArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].pdf(x[i])
        return retval

    def entropy(self) -> RealArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].entropy()
        return retval



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parameter import ScalarSupport, VectorSupport, distribution_parameter
from ...samplable import Samplable

__all__ = ['IsotropicNormalNP', 'IsotropicNormalEP']


@dataclass
class IsotropicNormalNP(NaturalParametrization['IsotropicNormalEP', RealArray]):
    mean_times_precision: RealArray = distribution_parameter(VectorSupport())
    negative_half_precision: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.negative_half_precision.shape

    def log_normalizer(self) -> RealArray:
        eta = self.mean_times_precision
        num_parameters = eta.shape[-1]
        return 0.5 * (-0.5 * jnp.sum(jnp.square(eta), axis=-1) / self.negative_half_precision
                      + num_parameters * jnp.log(jnp.pi / -self.negative_half_precision))

    def to_exp(self) -> IsotropicNormalEP:
        num_parameters = self.mean_times_precision.shape[-1]
        precision = -2.0 * self.negative_half_precision
        mean = self.mean_times_precision / precision[..., jnp.newaxis]
        total_variance = num_parameters / precision
        total_second_moment = jnp.sum(jnp.square(mean), axis=-1) + total_variance
        return IsotropicNormalEP(mean, total_second_moment)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape[:-1])

    def sufficient_statistics(self, x: RealArray) -> IsotropicNormalEP:
        return IsotropicNormalEP(x, jnp.sum(jnp.square(x), axis=-1))

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class IsotropicNormalEP(ExpectationParametrization[IsotropicNormalNP], Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())
    total_second_moment: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[IsotropicNormalNP]:
        return IsotropicNormalNP

    def to_nat(self) -> IsotropicNormalNP:
        variance = self.variance()
        negative_half_precision = -0.5 / variance
        mean_times_precision = self.mean / variance[..., jnp.newaxis]
        return IsotropicNormalNP(mean_times_precision, negative_half_precision)

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        deviation = jnp.sqrt(self.variance())
        return jax.random.normal(rng.key, shape) * deviation + self.mean

    # New methods ----------------------------------------------------------------------------------
    def variance(self) -> RealArray:
        num_parameters = self.mean.shape[-1]
        return (self.total_second_moment - jnp.sum(jnp.square(self.mean), axis=-1)) / num_parameters

    def dimensions(self) -> int:
        return self.mean.shape[-1]



from .bernoulli import *
from .beta import *
from .chi import *
from .chi_square import *
from .cmvn import *
from .complex_normal import *
from .dirichlet import *
from .exponential import *
from .gamma import *
from .geometric import *
from .logarithmic import *
from .multinomial import *
from .multivariate_normal import *
from .negative_binomial import *
from .normal import *
from .poisson import *
from .rayleigh import *
from .von_mises import *
from .weibull import *



from __future__ import annotations

import math
from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import ComplexArray, Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parameter import SymmetricMatrixSupport, distribution_parameter
from ...samplable import Samplable

__all__ = ['ComplexCircularlySymmetricNormalNP', 'ComplexCircularlySymmetricNormalEP']


def _broadcasted_outer_c(x: ComplexArray) -> ComplexArray:
    return jnp.einsum("...i,...j->...ij", x, x.conjugate())


@dataclass
class ComplexCircularlySymmetricNormalNP(
        NaturalParametrization['ComplexCircularlySymmetricNormalEP', ComplexArray],
        Samplable):
    """
    The complex multivariate normal distribution with zero mean and and zero pseudo-variance.  This
    is a curved exponential family.
    """
    negative_precision: ComplexArray = distribution_parameter(
        SymmetricMatrixSupport(hermitian=True))
    # S = -1/negative_precision, U = 0
    # P = S.conjugate, R = 0
    # H = -1/S, J = 0
    # K = 0, L = S/2
    # eta = 0
    # Leta = 0

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.negative_precision.shape[:-2]

    def log_normalizer(self) -> RealArray:
        num_parameters = self.negative_precision.shape[-1]
        log_det_s = jnp.log(jnp.linalg.det(-self.negative_precision).real)
        return -log_det_s + num_parameters * math.log(math.pi)

    def to_exp(self) -> ComplexCircularlySymmetricNormalEP:
        return ComplexCircularlySymmetricNormalEP(
            jnp.linalg.inv(-self.negative_precision).conjugate())

    def carrier_measure(self, x: ComplexArray) -> RealArray:
        return jnp.zeros(x.shape[:-1])

    def sufficient_statistics(self, x: ComplexArray) -> ComplexCircularlySymmetricNormalEP:
        return ComplexCircularlySymmetricNormalEP(_broadcasted_outer_c(x))

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> ComplexArray:
        return self.to_exp().sample(rng, shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.negative_precision.shape[-1]


@dataclass
class ComplexCircularlySymmetricNormalEP(
        ExpectationParametrization[ComplexCircularlySymmetricNormalNP], Samplable):
    variance: ComplexArray = distribution_parameter(SymmetricMatrixSupport(hermitian=True))

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.variance.shape[:-2]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[ComplexCircularlySymmetricNormalNP]:
        return ComplexCircularlySymmetricNormalNP

    def to_nat(self) -> ComplexCircularlySymmetricNormalNP:
        return ComplexCircularlySymmetricNormalNP(
            -jnp.linalg.inv(self.variance).conjugate())

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> ComplexArray:
        if shape is not None:
            shape += self.variance.shape[:-1]
        else:
            shape = self.variance.shape[:-1]
        mean = jnp.zeros(shape)
        return jax.random.multivariate_normal(rng.key, mean, self.variance,
                                              shape[:-1])  # type: ignore

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.variance.shape[-1]



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from tjax import RealArray
from tjax.dataclasses import dataclass

from ..parameter import ScalarSupport, distribution_parameter
from .negative_binomial_common import NBCommonEP, NBCommonNP

__all__ = ['GeometricNP', 'GeometricEP']


@dataclass
class GeometricNP(NBCommonNP['GeometricEP']):
    log_not_p: RealArray = distribution_parameter(ScalarSupport())
    failures = 1

    # Implemented methods --------------------------------------------------------------------------
    def to_exp(self) -> GeometricEP:
        return GeometricEP(self._mean())

    def sufficient_statistics(self, x: RealArray) -> GeometricEP:
        return GeometricEP(x)

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.log_not_p.shape)


@dataclass
class GeometricEP(NBCommonEP[GeometricNP]):
    mean: RealArray = distribution_parameter(ScalarSupport())
    failures = 1

    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[GeometricNP]:
        return GeometricNP

    def to_nat(self) -> GeometricNP:
        return GeometricNP(self._log_not_p())



from __future__ import annotations

from numbers import Real
from typing import Optional, Type, TypeVar

import numpy as np
from numpy.random import Generator
from tjax import ComplexArray, ComplexNumeric, RealArray, RealNumeric, ShapeLike

from .multivariate_normal import ScipyMultivariateNormal, ScipyMultivariateNormalUnvectorized
from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyComplexNormal']


class ScipyComplexNormalUnvectorized:
    """
    Represents an array of univariate complex normal distributions.
    """
    def __init__(self,
                 mean: ComplexNumeric,
                 variance: RealNumeric,
                 pseudo_variance: ComplexNumeric):
        self.mean: ComplexArray = np.asarray(mean)
        self.variance: RealArray = np.asarray(variance)
        self.pseudo_variance: ComplexArray = np.asarray(pseudo_variance)
        if not issubclass(self.variance.dtype.type, Real):
            raise TypeError(f"The variance {self.variance} has non-real "
                            f"dtype {self.variance.dtype}.")
        if np.any(np.abs(self.pseudo_variance) > self.variance):
            raise ValueError(f"The pseudo-variance {pseudo_variance} "
                             f"is bigger than the variance {variance}.")
        if not (self.mean.shape == self.variance.shape
                == self.pseudo_variance.shape):
            raise ValueError("Shape mismatch.")

    # New methods ----------------------------------------------------------------------------------
    def pdf(self, z: ComplexNumeric, out: None = None) -> RealNumeric:
        zr = np.stack([z.real, z.imag], axis=-1)
        return self.as_multivariate_normal().pdf(zr)

    def rvs(self, size: ShapeLike = (), random_state: Optional[Generator] = None) -> ComplexArray:
        if isinstance(size, int):
            size = (size,)
        if random_state is None:
            random_state = np.random.default_rng()
        xy_rvs = random_state.multivariate_normal(mean=self._multivariate_normal_mean(),
                                                  cov=self._multivariate_normal_cov(),
                                                  size=size)
        return xy_rvs[..., 0] + 1j * xy_rvs[..., 1]

    def entropy(self) -> RealArray:
        return self.as_multivariate_normal().entropy()

    def as_multivariate_normal(self) -> ScipyMultivariateNormalUnvectorized:
        mv_mean = self._multivariate_normal_mean()
        mv_cov = self._multivariate_normal_cov()
        return ScipyMultivariateNormalUnvectorized(mean=mv_mean, cov=mv_cov)

    # Private methods ------------------------------------------------------------------------------
    def _multivariate_normal_mean(self) -> RealArray:
        return np.stack([self.mean.real, self.mean.imag], axis=-1)

    def _multivariate_normal_cov(self) -> RealArray:
        cov_sum = self.variance + self.pseudo_variance
        cov_diff = self.variance - self.pseudo_variance
        xx = 0.5 * cov_sum.real
        xy = 0.5 * -cov_diff.imag
        yx = 0.5 * cov_sum.imag
        yy = 0.5 * cov_diff.real
        xx_xy = np.stack([xx, xy], axis=-1)
        yx_yy = np.stack([yx, yy], axis=-1)
        return np.stack([xx_xy, yx_yy], axis=-2)


class ScipyComplexNormal(ShapedDistribution):
    """
    This class allows distributions having a non-empty shape.
    """
    def __init__(self,
                 mean: Optional[ComplexArray] = None,
                 variance: Optional[RealArray] = None,
                 pseudo_variance: Optional[ComplexArray] = None):
        if mean is not None:
            shape = mean.shape
        elif variance is not None:
            shape = variance.shape
        elif pseudo_variance is not None:
            shape = pseudo_variance.shape
        else:
            raise ValueError
        dtype = np.result_type(*[x.dtype  # type: ignore
                                 for x in [mean, variance, pseudo_variance]
                                 if x is not None])
        rvs_shape = ()
        if mean is None:
            mean = np.zeros(shape, dtype=dtype)
        if variance is None:
            variance = np.array(np.ones(shape, dtype=dtype).real)
        if pseudo_variance is None:
            pseudo_variance = np.zeros(shape, dtype=dtype)

        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ScipyComplexNormalUnvectorized(mean[i], variance[i], pseudo_variance[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    _T = TypeVar('_T', bound='ScipyComplexNormal')

    @classmethod
    def init_using_angle(cls: Type[_T],
                         mean: ComplexNumeric,
                         variance: RealNumeric,
                         angle: RealNumeric,
                         polarization: ComplexNumeric) -> _T:
        r = polarization * np.exp(1j * 2 * np.pi * angle * 2)
        mean_array = np.array(mean)
        variance_array = np.array(variance)
        pseudo_variance_array = np.array(r * variance)
        return cls(mean_array, variance_array, pseudo_variance_array)

    def as_multivariate_normal(self) -> ScipyMultivariateNormal:
        objects = np.empty(self.shape, dtype=np.object_)
        for i in np.ndindex(*self.shape):
            objects[i] = self.objects[i].as_multivariate_normal()
        return ScipyMultivariateNormal(self.shape, self.rvs_shape, self.real_dtype, objects)

    @property
    def mean(self) -> ComplexArray:
        retval = np.empty(self.shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].mean
        return retval

    @property
    def variance(self) -> ComplexArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].variance
        return retval

    @property
    def pseudo_variance(self) -> ComplexArray:
        retval = np.empty(self.shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].pseudo_variance
        return retval



from typing import Optional

import numpy as np
import scipy.stats as ss
from tjax import RealArray

from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyVonMises']


class ScipyVonMises(ShapedDistribution):
    """
    This class allows distributions having a non-empty shape.
    """
    def __init__(self, kappa: RealArray, loc: Optional[RealArray] = None):
        if loc is None:
            loc = np.zeros_like(kappa)
        shape = np.broadcast(kappa, loc).shape
        rvs_shape = ()
        dtype = np.result_type(kappa.dtype, loc.dtype)  # type: ignore
        kappa = np.broadcast_to(kappa, shape).astype(dtype)  # type: ignore
        loc: RealArray = np.broadcast_to(loc, shape).astype(dtype)  # type: ignore
        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ss.vonmises(kappa[i], loc[i])
        super().__init__(shape, rvs_shape, dtype, objects)



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..conjugate_prior import HasConjugatePrior
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable
from .gamma import GammaNP

__all__ = ['ExponentialNP', 'ExponentialEP']


@dataclass
class ExponentialNP(NaturalParametrization['ExponentialEP', RealArray], Samplable):
    negative_rate: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.negative_rate.shape

    def log_normalizer(self) -> RealArray:
        return -jnp.log(-self.negative_rate)

    def to_exp(self) -> ExponentialEP:
        return ExponentialEP(-1.0 / self.negative_rate)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sufficient_statistics(self, x: RealArray) -> ExponentialEP:
        return ExponentialEP(x)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return -jax.random.exponential(rng.key, shape) / self.negative_rate


@dataclass
class ExponentialEP(HasConjugatePrior[ExponentialNP], Samplable):
    mean: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[ExponentialNP]:
        return ExponentialNP

    def to_nat(self) -> ExponentialNP:
        return ExponentialNP(-1.0 / self.mean)

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.exponential(rng.key, shape) * self.mean

    # Overridden methods ---------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> GammaNP:
        return GammaNP(-n / self.mean, n)

    def conjugate_prior_observation(self) -> RealArray:
        return self.mean



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
from jax.nn import softplus
from jax.scipy import special as jss
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ..exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter
from ..samplable import Samplable
from ..tools import inverse_softplus

__all__ = ['GammaNP', 'GammaEP']


@dataclass
class GammaNP(NaturalParametrization['GammaEP', RealArray], Samplable):
    negative_rate: RealArray = distribution_parameter(ScalarSupport())
    shape_minus_one: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.negative_rate.shape

    def log_normalizer(self) -> RealArray:
        shape = self.shape_minus_one + 1.0
        return jss.gammaln(shape) - shape * jnp.log(-self.negative_rate)

    def to_exp(self) -> GammaEP:
        shape = self.shape_minus_one + 1.0
        return GammaEP(-shape / self.negative_rate,
                       jss.digamma(shape) - jnp.log(-self.negative_rate))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape)

    def sufficient_statistics(self, x: RealArray) -> GammaEP:
        return GammaEP(x, jnp.log(x))

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        return -jax.random.gamma(rng.key, self.shape_minus_one + 1.0, shape) / self.negative_rate


@dataclass
class GammaEP(ExpToNat[GammaNP, RealArray]):
    mean: RealArray = distribution_parameter(ScalarSupport())
    mean_log: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape

    @classmethod
    def natural_parametrization_cls(cls) -> Type[GammaNP]:
        return GammaNP

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def search_to_natural(self, search_parameters: RealArray) -> GammaNP:
        shape = softplus(search_parameters)
        rate = shape / self.mean
        return GammaNP(-rate, shape - 1.0)

    def search_gradient(self, search_parameters: RealArray) -> RealArray:
        shape = softplus(search_parameters)
        log_mean_minus_mean_log = jnp.log(self.mean) - self.mean_log
        return log_mean_minus_mean_log - jnp.log(shape) + jss.digamma(shape)
        # gradient is 1.0 / shape - jss.polygamma(1, shape)
        # where polygamma(1) is trigamma

    # Overridden methods ---------------------------------------------------------------------------
    def initial_search_parameters(self) -> RealArray:
        log_mean_minus_mean_log = jnp.log(self.mean) - self.mean_log
        initial_shape = (
            (3.0 - log_mean_minus_mean_log
             + jnp.sqrt((log_mean_minus_mean_log - 3.0) ** 2 + 24.0 * log_mean_minus_mean_log))
            / (12.0 * log_mean_minus_mean_log))
        return inverse_softplus(initial_shape)



from typing import Any, Generic

from tjax import ComplexArray, RealArray

from .expectation_parametrization import NP, ExpectationParametrization
from .natural_parametrization import NaturalParametrization

__all__ = ['HasConjugatePrior']


class HasConjugatePrior(ExpectationParametrization[NP], Generic[NP]):
    # Abstract methods -----------------------------------------------------------------------------
    def conjugate_prior_distribution(self, n: RealArray) -> NaturalParametrization[Any, Any]:
        """
        Args:
            n: The nonnegative number of pseudo-observations.  Must have shape == self.shape.
        """
        raise NotImplementedError

    def conjugate_prior_observation(self) -> ComplexArray:
        raise NotImplementedError



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from jax.scipy import special as jss
from tjax import RealArray, Shape
from tjax.dataclasses import dataclass

from ..exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter

__all__ = ['ChiSquareNP', 'ChiSquareEP']


@dataclass
class ChiSquareNP(NaturalParametrization['ChiSquareEP', RealArray]):
    """
    The chi-square distribution with k degrees of freedom is the gamma distribution with shape k/2
    and rate 1/2.
    """
    k_over_two_minus_one: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.k_over_two_minus_one.shape

    def log_normalizer(self) -> RealArray:
        k_over_two = self.k_over_two_minus_one + 1.0
        return jss.gammaln(k_over_two) - k_over_two * jnp.log(0.5)

    def to_exp(self) -> ChiSquareEP:
        k_over_two = self.k_over_two_minus_one + 1.0
        return ChiSquareEP(jss.digamma(k_over_two) - jnp.log(0.5))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return -x * 0.5

    def sufficient_statistics(self, x: RealArray) -> ChiSquareEP:
        return ChiSquareEP(jnp.log(x))


# The ExpToNat mixin can be circumvented if the inverse of the digamma function were added to JAX.
@dataclass
class ChiSquareEP(ExpToNat[ChiSquareNP, ChiSquareNP]):
    mean_log: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[ChiSquareNP]:
        return ChiSquareNP

    @property
    def shape(self) -> Shape:
        return self.mean_log.shape

    def expected_carrier_measure(self) -> RealArray:
        q = self.to_nat()
        k_over_two = q.k_over_two_minus_one + 1.0
        return -k_over_two

    def initial_search_parameters(self) -> ChiSquareNP:
        return ChiSquareNP(jnp.zeros(self.mean_log.shape))

    def search_to_natural(self, search_parameters: ChiSquareNP) -> ChiSquareNP:
        return search_parameters

    def search_gradient(self, search_parameters: ChiSquareNP) -> ChiSquareNP:
        return self._natural_gradient(search_parameters)



from typing import Optional

import numpy as np
import scipy.stats as ss
from numpy.random import Generator
from tjax import ComplexArray, RealArray, ShapeLike

from .shaped_distribution import ShapedDistribution

__all__ = ['ScipyDirichlet']


# pylint: disable=protected-access
class ScipyDirichletFixRVsAndPDF(ss._multivariate.dirichlet_frozen):
    """
    This class repairs dirichlet.  See https://github.com/scipy/scipy/issues/6005 and
    https://github.com/scipy/scipy/issues/6006.
    """
    def rvs(self,
            size: Optional[ShapeLike] = None,
            random_state: Optional[Generator] = None) -> RealArray:
        if size is None:
            size = ()
        # This somehow fixes the behaviour of rvs.
        return super().rvs(size=size, random_state=random_state)

    def pdf(self, x: RealArray) -> RealArray:
        if x.ndim == 2:
            return super().pdf(x.T)
        return super().pdf(x)


class ScipyDirichlet(ShapedDistribution):
    """
    This class allows distributions having a non-empty shape.
    """
    def __init__(self, alpha: RealArray):
        shape = alpha[..., -1].shape
        rvs_shape = (alpha.shape[-1],)
        dtype = alpha.dtype
        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(shape):
            objects[i] = ScipyDirichletFixRVsAndPDF(alpha[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    def pdf(self, x: ComplexArray) -> RealArray:
        x = x.astype(np.float64)
        y = np.sum(x, axis=-1)
        if not np.allclose(y, np.ones(y.shape), atol=1e-5, rtol=0):
            raise ValueError
        return super().pdf(x)



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from tjax import RealArray
from tjax.dataclasses import dataclass

from .dirichlet_common import DirichletCommonEP, DirichletCommonNP

__all__ = ['DirichletNP', 'DirichletEP']


@dataclass
class DirichletNP(DirichletCommonNP['DirichletEP']):
    # Implemented methods --------------------------------------------------------------------------
    def to_exp(self) -> DirichletEP:
        return DirichletEP(self._exp_helper())

    def sufficient_statistics(self, x: RealArray) -> DirichletEP:
        one_minus_total_x = 1.0 - jnp.sum(x, axis=-1, keepdims=True)
        return DirichletEP(jnp.append(jnp.log(x), jnp.log(one_minus_total_x), axis=-1))

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape[:len(x.shape) - 1])


@dataclass
class DirichletEP(DirichletCommonEP[DirichletNP]):
    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[DirichletNP]:
        return DirichletNP

    def search_to_natural(self, search_parameters: RealArray) -> DirichletNP:
        return DirichletNP(self._transform_nat_helper(search_parameters))



from __future__ import annotations

from typing import Optional, Type

import jax
import jax.numpy as jnp
import numpy as np
from tjax import Generator, RealArray, Shape
from tjax.dataclasses import dataclass

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parameter import SymmetricMatrixSupport, VectorSupport, distribution_parameter
from ...samplable import Samplable

__all__ = ['MultivariateNormalNP', 'MultivariateNormalEP', 'MultivariateNormalVP']


def _broadcasted_outer(x: RealArray) -> RealArray:
    return jnp.einsum("...i,...j->...ij", x, x)


@dataclass
class MultivariateNormalNP(NaturalParametrization['MultivariateNormalEP', RealArray]):
    mean_times_precision: RealArray = distribution_parameter(VectorSupport())
    negative_half_precision: RealArray = distribution_parameter(SymmetricMatrixSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    def log_normalizer(self) -> RealArray:
        eta = self.mean_times_precision
        k = eta.shape[-1]
        h_inv = jnp.linalg.inv(self.negative_half_precision)
        a = jnp.einsum("...i,...ij,...j", eta, h_inv, eta)
        _, ld = jnp.linalg.slogdet(-self.negative_half_precision)
        return -0.25 * a - 0.5 * ld + 0.5 * k * jnp.log(np.pi)

    def to_exp(self) -> MultivariateNormalEP:
        h_inv = jnp.linalg.inv(self.negative_half_precision)
        h_inv_times_eta = jnp.einsum("...ij,...j->...i", h_inv, self.mean_times_precision)
        mean = -0.5 * h_inv_times_eta
        second_moment = 0.25 * _broadcasted_outer(h_inv_times_eta) - 0.5 * h_inv
        return MultivariateNormalEP(mean, second_moment)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.zeros(x.shape[:-1])

    def sufficient_statistics(self, x: RealArray) -> MultivariateNormalEP:
        return MultivariateNormalEP(x, _broadcasted_outer(x))

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateNormalEP(ExpectationParametrization[MultivariateNormalNP], Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())
    second_moment: RealArray = distribution_parameter(SymmetricMatrixSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    def to_nat(self) -> MultivariateNormalNP:
        precision = jnp.linalg.inv(self.variance())
        mean_times_precision = jnp.einsum("...ij,...j->...i", precision, self.mean)
        return MultivariateNormalNP(mean_times_precision, -0.5 * precision)

    def expected_carrier_measure(self) -> RealArray:
        return jnp.zeros(self.shape)

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        return self.to_variance_parametrization().sample(rng, shape)

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> RealArray:
        return self.second_moment - _broadcasted_outer(self.mean)

    def to_variance_parametrization(self) -> MultivariateNormalVP:
        return MultivariateNormalVP(self.mean, self.variance())


@dataclass
class MultivariateNormalVP(Samplable):
    mean: RealArray = distribution_parameter(VectorSupport())
    variance: RealArray = distribution_parameter(SymmetricMatrixSupport())

    # Implemented methods --------------------------------------------------------------------------
    @property
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @classmethod
    def natural_parametrization_cls(cls) -> Type[MultivariateNormalNP]:
        return MultivariateNormalNP

    def sample(self, rng: Generator, shape: Optional[Shape] = None) -> RealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.multivariate_normal(rng.key, self.mean, self.variance,
                                              shape)  # type: ignore

    # New methods ----------------------------------------------------------------------------------
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def to_exp(self) -> MultivariateNormalEP:
        second_moment = self.variance + _broadcasted_outer(self.mean)
        return MultivariateNormalEP(self.mean, second_moment)



from __future__ import annotations

from typing import Any, Generic, List, Tuple, TypeVar

import jax.numpy as jnp
from jax import jit
from jax.tree_util import tree_map
from tjax.dataclasses import dataclass, field
from tjax.fixed_point import ComparingIteratedFunctionWithCombinator, ComparingState
from tjax.gradient import Adam, GradientTransformation

from .expectation_parametrization import ExpectationParametrization
from .natural_parametrization import NaturalParametrization

__all__: List[str] = []


NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])
SP = TypeVar('SP', bound=Any)


class ExpToNat(ExpectationParametrization[NP], Generic[NP, SP]):
    """
    This mixin implements the conversion from expectation to natural parameters using Newton's
    method with a Jacobian to invert the gradient log-normalizer.
    """

    # Implemented methods --------------------------------------------------------------------------
    @jit
    def to_nat(self) -> NP:
        iterated_function = ExpToNatIteratedFunction[NP, SP](minimum_iterations=1000,
                                                             maximum_iterations=1000,
                                                             transform=Adam(1e-1))
        initial_search_parameters = self.initial_search_parameters()
        initial_gt_state = iterated_function.transform.init(initial_search_parameters)
        initial_state = initial_gt_state, initial_search_parameters
        final_state = iterated_function.find_fixed_point(self, initial_state).current_state
        _, final_search_parameters = final_state
        return self.search_to_natural(final_search_parameters)

    # Non-final methods ----------------------------------------------------------------------------
    def _zero_natural_parameters(self) -> NP:
        "A convenience method for implementing initial_search_parameters"
        fixed_parameters = self.fixed_parameters_mapping()
        cls = type(self).natural_parametrization_cls()
        return cls.unflattened(jnp.zeros_like(self.flattened()), **fixed_parameters)

    # Abstract methods -----------------------------------------------------------------------------
    def initial_search_parameters(self) -> SP:
        """
        Returns: The initial value of the parameters that Newton's method runs on.
        """
        raise NotImplementedError

    def search_to_natural(self, search_parameters: SP) -> NP:
        """
        Args:
            search_parameters: The parameters that Newton's method runs on.
        Returns: The corresponding natural parametrization.
        """
        raise NotImplementedError

    def search_gradient(self, search_parameters: SP) -> SP:
        raise NotImplementedError

    # Private methods ------------------------------------------------------------------------------
    def _natural_gradient(self, natural_parameters: NP) -> NP:
        """
        Returns: The difference of the expectation parameters corresponding to natural_parameters
            and self.  This difference is returned as natural parameters.
        """
        expectation_parameters: ExpToNat[NP, SP] = natural_parameters.to_exp()
        exp_difference: ExpToNat[NP, SP] = tree_map(jnp.subtract, expectation_parameters, self)
        return type(natural_parameters).unflattened(exp_difference.flattened())


@dataclass
class ExpToNatIteratedFunction(
        # The generic parameters are: Parameters, State, Comparand, Differentiand, Trajectory.
        ComparingIteratedFunctionWithCombinator[ExpToNat[NP, SP],
                                                Tuple[Any, SP],
                                                SP,
                                                SP,
                                                NP],
        Generic[NP, SP]):

    transform: GradientTransformation[Any, NP] = field()  # TODO: kw_only=True

    def sampled_state(self, theta: ExpToNat[NP, SP], state: Tuple[Any, SP]) -> Tuple[Any, SP]:
        current_gt_state, search_parameters = state
        search_gradient = theta.search_gradient(search_parameters)
        transformed_gradient, new_gt_state = self.transform.update(search_gradient,
                                                                   current_gt_state,
                                                                   search_parameters)
        new_search_parameters = tree_map(jnp.add, search_parameters, transformed_gradient)
        return new_gt_state, new_search_parameters

    def sampled_state_trajectory(
            self,
            theta: ExpToNat[NP, SP],
            augmented: ComparingState[Tuple[Any, SP], SP]) -> Tuple[Tuple[Any, SP], NP]:
        sampled_state = self.sampled_state(theta, augmented.current_state)
        _, trajectory = sampled_state
        return sampled_state, trajectory

    def extract_comparand(self, state: Tuple[Any, SP]) -> SP:
        _, search_parameters = state
        return search_parameters

    def extract_differentiand(self,
                              theta: ExpToNat[NP, SP],
                              state: Tuple[Any, SP]) -> SP:
        _, search_parameters = state
        return search_parameters

    def implant_differentiand(self,
                              theta: ExpToNat[NP, SP],
                              state: Tuple[Any, SP],
                              differentiand: SP) -> Tuple[Any, SP]:
        current_gt_state, _ = state
        return current_gt_state, differentiand



from __future__ import annotations

from typing import Type

import jax.numpy as jnp
from jax.scipy import special as jss
from tjax import Array, RealArray
from tjax.dataclasses import dataclass

from ..parameter import ScalarSupport, distribution_parameter
from ..transformed_parametrization import (TransformedExpectationParametrization,
                                           TransformedNaturalParametrization)
from .chi_square import ChiSquareEP, ChiSquareNP

__all__ = ['ChiNP', 'ChiEP']


@dataclass
class ChiNP(TransformedNaturalParametrization[ChiSquareNP, ChiSquareEP, 'ChiEP', RealArray]):
    k_over_two_minus_one: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    def base_distribution(self) -> ChiSquareNP:
        return ChiSquareNP(self.k_over_two_minus_one)

    def create_expectation(self, expectation_parametrization: ChiSquareEP) -> ChiEP:
        return ChiEP(expectation_parametrization.mean_log)

    def sample_to_base_sample(self, x: Array) -> RealArray:
        return jnp.square(x)

    def carrier_measure(self, x: RealArray) -> RealArray:
        return jnp.log(2.0 * x) - jnp.square(x) * 0.5


@dataclass
class ChiEP(TransformedExpectationParametrization[ChiSquareEP, ChiSquareNP, ChiNP]):
    mean_log: RealArray = distribution_parameter(ScalarSupport())

    # Implemented methods --------------------------------------------------------------------------
    @classmethod
    def natural_parametrization_cls(cls) -> Type[ChiNP]:
        return ChiNP

    def base_distribution(self) -> ChiSquareEP:
        return ChiSquareEP(self.mean_log)

    def create_natural(self, natural_parametrization: ChiSquareNP) -> ChiNP:
        return ChiNP(natural_parametrization.k_over_two_minus_one)

    def expected_carrier_measure(self) -> RealArray:
        q = self.to_nat()
        k_over_two = q.k_over_two_minus_one + 1.0
        return -1.0 * k_over_two + 0.5 * jss.digamma(k_over_two) + 1.5 * jnp.log(2.0)



from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, Any, Iterable, List

import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from tjax import ComplexArray, ComplexNumeric, RealArray, RealNumeric

__all__: List[str] = []


def np_abs_square(x: ComplexNumeric) -> RealNumeric:
    return np.square(x.real) + np.square(x.imag)


def parameters_dot_product(x: NaturalParametrization[Any, Any], y: Any) -> RealArray:
    def dotted_fields() -> Iterable[RealArray]:
        for (x_value, x_support), (y_value, y_support) in zip(x.parameters_value_support(),
                                                              y.parameters_value_support()):
            axes = x_support.axes()
            assert y_support.axes() == axes
            yield _parameter_dot_product(x_value, y_value, axes)
    return reduce(jnp.add, dotted_fields())


def inverse_softplus(y: RealArray) -> RealArray:
    return jnp.where(y > 80.0,
                     y,
                     jnp.log(jnp.expm1(y)))


ive = tfp.math.bessel_ive
log_ive = tfp.math.log_bessel_ive


def iv(v: RealNumeric, z: RealNumeric) -> RealNumeric:
    return tfp.math.bessel_ive(v, z) / jnp.exp(-jnp.abs(z))


def vectorized_tril(m: RealArray, k: int = 0) -> RealArray:
    n, m_ = m.shape[-2:]
    indices = (..., *np.tril_indices(n, k, m_))  # type: ignore
    values = m[indices]
    retval = np.zeros_like(m)
    retval[indices] = values
    return retval


def vectorized_triu(m: RealArray, k: int = 0) -> RealArray:
    n, m_ = m.shape[-2:]
    indices = (..., *np.triu_indices(n, k, m_))  # type: ignore
    values = m[indices]
    retval = np.zeros_like(m)
    retval[indices] = values
    return retval


def create_diagonal(m: RealArray) -> RealArray:
    """
    Args:
        m: Has shape (*k, n)
    Returns: Array with shape (*k, n, n) and the elements of m on the diagonals.
    """
    indices = (..., *np.diag_indices(m.shape[-1]))
    retval = np.zeros(m.shape + (m.shape[-1],), dtype=m.dtype)
    retval[indices] = m
    return retval


# Private functions --------------------------------------------------------------------------------
def _parameter_dot_product(x: ComplexArray, y: ComplexArray, n_axes: int) -> RealArray:
    """
    Returns the real component of the dot product of the final n_axes axes of two arrays.
    """
    axes = tuple(range(-n_axes, 0))
    return jnp.sum(x * y, axis=axes).real


if TYPE_CHECKING:
    from .natural_parametrization import NaturalParametrization
